/* This file is part of the KDE libraries
    Copyright (C) 1999,2000 Stephan Kulow <coolo@kde.org>
                  1999,2000 Carsten Pfeiffer <pfeiffer@kde.org>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include <unistd.h>

#include <qdir.h>
#include <qapplication.h>
#include <qdialog.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qpopupmenu.h>
#include <qtimer.h>

#include <kaction.h>
#include <kapp.h>
#include <kdebug.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kprogress.h>
#include <kstdaction.h>

#include <kcombiview.h>
#include "diroperator.h"
//#include <kfiledetailview.h>
#include <kfilednddetailview.h>
//#include <kfileiconview.h>
#include <kfiledndiconview.h>
#include <kfilepreview.h>
//#include <kfilereader.h>
#include <kfileview.h>
#include <kfileviewitem.h>
#include <kprocess.h>
#include "inputbox.h"
#include "config-kfile.h"


template class QStack<KURL>;
template class QDict<KFileViewItem>;
#define MAIL

KURL *DirOperator::lastDirectory = 0; // to set the start path

class DirOperator::DirOperatorPrivate
{
public:
    bool onlyDoubleClickSelectsFiles;
};

DirOperator::DirOperator(const KURL& url,
                           QWidget *parent, const char* _name)
    : QWidget(parent, _name), fileView(0), oldView(0), progress(0)
{
///////////////////////////////	
	  setAcceptDrops(TRUE);
//    tmpFile = "";
///////////////////////////////
    myPreview = 0L;
    myMode = KFile::File;
    viewKind = KFile::Simple;
    mySorting = static_cast<QDir::SortSpec>(QDir::Name | QDir::DirsFirst);
    d = new DirOperatorPrivate;
    d->onlyDoubleClickSelectsFiles = true;

    if (url.isEmpty()) { // no dir specified -> current dir
        QString strPath = QDir::currentDirPath();
        strPath.append('/');
        lastDirectory = new KURL;
        lastDirectory->setProtocol(QString::fromLatin1("file"));
        lastDirectory->setPath(strPath);
    }
    else {
        lastDirectory = new KURL(url);
        if ( lastDirectory->protocol().isEmpty() )
            lastDirectory->setProtocol(QString::fromLatin1("file"));
    }

    dir = new FileReader(*lastDirectory);
    dir->setAutoUpdate( true );

    connect(dir, SIGNAL(newItems(const KFileItemList &)), SLOT(insertNewFiles(const KFileItemList &)));
    connect(dir, SIGNAL(completed()), SLOT(slotIOFinished()));
    connect(dir, SIGNAL(canceled()), SLOT(resetCursor()));
    connect(dir, SIGNAL(deleteItem(KFileItem *)), SLOT(itemDeleted(KFileItem *)));
    connect(&myCompletion, SIGNAL(match(const QString&)), SLOT(slotCompletionMatch(const QString&)));

    finished = true;
    myCompleteListDirty = false;

    backStack.setAutoDelete( true );
    forwardStack.setAutoDelete( true );

    // action stuff
    setupActions();
    setupMenu();

    setFocusPolicy(QWidget::WheelFocus);
}

DirOperator::~DirOperator()
{
    resetCursor();
    delete fileView;
    delete dir;
    delete d;
}


void DirOperator::setSorting( QDir::SortSpec spec )
{
    if ( fileView )
        fileView->setSorting( spec );
    mySorting = spec;
    updateSortActions();
}

void DirOperator::readNextMimeType()
{
    if (pendingMimeTypes.isEmpty()) {
        return;
    }

    KFileViewItem *item = pendingMimeTypes.first();
    const QPixmap& p = item->pixmap();
    (void) item->determineMimeType();

    if ( item->pixmap().serialNumber() != p.serialNumber() )
        fileView->updateView(item);
    pendingMimeTypes.removeFirst();
    QTimer::singleShot(0, this, SLOT(readNextMimeType()));
}

void DirOperator::resetCursor()
{
    if (!finished)
        QApplication::restoreOverrideCursor();
    finished = false;
    if (progress)
        progress->hide();
}

void DirOperator::activatedMenu( const KFileViewItem * )
{
//    QTimer::singleShot(0, this, SLOT(readNextMimeType()));
    actionMenu->popup( QCursor::pos() );
}
void DirOperator::mouseReleaseEvent(QMouseEvent *e)
{
//    actionMenu->popup( QCursor::pos() );
}

void DirOperator::setPreviewWidget(const QWidget *w) {

    if(w != 0L)
        viewKind = (viewKind | KFile::PreviewContents & ~KFile::SeparateDirs);
    else
        viewKind = (viewKind & ~KFile::PreviewContents);
    myPreview=const_cast<QWidget*>(w);
    setView( static_cast<KFile::FileView>(viewKind) );
}

int DirOperator::numDirs() const
{
    return fileView ? fileView->numDirs() : 0;
}

int DirOperator::numFiles() const
{
    return fileView ? fileView->numFiles() : 0;
}

void DirOperator::slotDetailedView()
{
    KFile::FileView view = static_cast<KFile::FileView>( viewKind & ~KFile::Simple | KFile::Detail );
    setView( view );
}

void DirOperator::slotSimpleView()
{
    KFile::FileView view = static_cast<KFile::FileView>( viewKind & ~KFile::Detail | KFile::Simple );
    setView( view );
}

void DirOperator::slotToggleHidden( bool show )
{
    dir->setShowingDotFiles( show );
    rereadDir();
}

void DirOperator::slotToggleMixDirsAndFiles()
{
    int flag = KFile::isSeparateDirs( (KFile::FileView) viewKind ) ? 0 : KFile::SeparateDirs;
    KFile::FileView view = static_cast<KFile::FileView>( viewKind & ~KFile::SeparateDirs | flag );
    setView( view );
}

void DirOperator::slotSortByName()
{
    int sorting = (fileView->sorting()) & ~QDir::SortByMask;
    fileView->setSorting( static_cast<QDir::SortSpec>( sorting | QDir::Name ));
    mySorting = fileView->sorting();
}

void DirOperator::slotSortBySize()
{
    int sorting = (fileView->sorting()) & ~QDir::SortByMask;
    fileView->setSorting( static_cast<QDir::SortSpec>( sorting | QDir::Size ));
    mySorting = fileView->sorting();
}

void DirOperator::slotSortByDate()
{
    int sorting = (fileView->sorting()) & ~QDir::SortByMask;
    fileView->setSorting( static_cast<QDir::SortSpec>( sorting | QDir::Time ));
    mySorting = fileView->sorting();
}

void DirOperator::slotSortReversed()
{
    if ( fileView )
        fileView->sortReversed();
}

void DirOperator::slotToggleDirsFirst()
{
    QDir::SortSpec sorting = fileView->sorting();
    if ( !KFile::isSortDirsFirst( sorting ) )
        fileView->setSorting( static_cast<QDir::SortSpec>( sorting | QDir::DirsFirst ));
    else
        fileView->setSorting( static_cast<QDir::SortSpec>( sorting & ~QDir::DirsFirst));
    mySorting = fileView->sorting();
}

void DirOperator::slotToggleIgnoreCase()
{
    QDir::SortSpec sorting = fileView->sorting();
    if ( !KFile::isSortCaseInsensitive( sorting ) )
        fileView->setSorting( static_cast<QDir::SortSpec>( sorting | QDir::IgnoreCase ));
    else
        fileView->setSorting( static_cast<QDir::SortSpec>( sorting & ~QDir::IgnoreCase));
    mySorting = fileView->sorting();
}

void DirOperator::mkdir()
{
    if (!dir->url().isLocalFile())
        return;

    // Modal widget asking the user the name of a new directory
    //
    QDialog *lMakeDir;
    QLabel *label;
    QLineEdit *ed;
    QVBoxLayout *lLayout;
    QPushButton *okButton, *cancelButton;
    QHBoxLayout *lLayout2;

    // Create widgets, and display using geometry management
    //
    lMakeDir = new QDialog(0,"MakeDir Dialog",true);
    lLayout = new QVBoxLayout( lMakeDir, 5 );
    label = new QLabel(lMakeDir);
    label->setAlignment( AlignLeft | AlignVCenter );
    label->setText(i18n("Create new directory in: ") + url().prettyURL() );
    label->setMinimumSize( label->sizeHint() );
    ed= new QLineEdit(lMakeDir);
    ed->setText( i18n("New Directory") );
    ed->adjustSize();
    ed->setFixedHeight( ed->height());
    ed->selectAll();
    connect(ed, SIGNAL(returnPressed()), lMakeDir, SLOT(accept()) );

    lLayout->addWidget( label, 0);
    lLayout->addSpacing( 5 );
    lLayout->addWidget( ed, 0);

    okButton = new QPushButton( lMakeDir, "okButton" );
    okButton->adjustSize();
    okButton->setText( i18n("&OK") );
    okButton->setFixedHeight( okButton->height() );
    okButton-> setMinimumWidth( okButton->width() );

    cancelButton = new QPushButton( lMakeDir, "cancelButton" );
    cancelButton->setText( i18n("&Cancel") );
    cancelButton->adjustSize();
    cancelButton->setFixedHeight( cancelButton->height() );
    cancelButton->setMinimumWidth( cancelButton->width() );

    connect( okButton, SIGNAL(clicked()), lMakeDir, SLOT(accept()) );
    connect( cancelButton, SIGNAL(clicked()), lMakeDir, SLOT(reject()) );

    lLayout2 = new QHBoxLayout( 5 );
    lLayout->addSpacing( 10 );
    lLayout->addLayout( lLayout2 , 0);
    lLayout2->addWidget( okButton, 3);
    lLayout2->addWidget( cancelButton, 3);
    lLayout->activate();

    // If the users presses enter (not escape) then create the dir
    // and insert it into the ListBox
    lMakeDir->resize( 10, 10);
    ed->grabKeyboard();
    if ( lMakeDir->exec() == QDialog::Accepted ) {

        // check if we are allowed to create directories
        bool writeOk = false;
        QString tmp = ed->text();
        int idx = tmp.findRev(QString::fromLatin1("/"));
        if ( idx != -1 ) {
            QDir dir( url().path() );
            if ( dir.cd( tmp.left( idx )) )
                writeOk = checkAccess( dir.absPath(), W_OK );
        }
        else
            writeOk = checkAccess( url().path(), W_OK );


        if ( !writeOk ) {
            KMessageBox::sorry(0,
                               i18n("You don't have permissions to create "
                                    "that directory." ));
            delete lMakeDir;
            return;
        }
        if ( QDir(url().path()).mkdir(ed->text()) == true ) {  // !! don't like this move it into FileReader ??
            setURL( KURL(url(), ed->text()), true );
        }
    }

    delete lMakeDir;
}

void DirOperator::close()
{
    resetCursor();
    pendingMimeTypes.clear();
    myCompletion.clear();
    myDirCompletion.clear();
}

void DirOperator::checkPath(const QString &, bool /*takeFiles*/) // SLOT
{
#if 0
    debug("KCOMMANDER2");
    // copy the argument in a temporary string
    QString text = _txt;
    // it's unlikely to happen, that at the beginning are spaces, but
    // for the end, it happens quite often, I guess.
    text = text.stripWhiteSpace();
    // if the argument is no URL (the check is quite fragil) and it's
    // no absolut path, we add the current directory to get a correct
    // url
    if (text.find(':') < 0 && text[0] != '/')
        text.insert(0, dir->url());

    // in case we have a selection defined and someone patched the file-
    // name, we check, if the end of the new name is changed.
    if (!selection.isNull()) {
        int position = text.findRev('/');
        ASSERT(position >= 0); // we already inserted the current dir in case
        QString filename = text.mid(position + 1, text.length());
        if (filename != selection)
            selection = QString::null;
    }

    KURL u(text); // I have to take care of entered URLs
    bool filenameEntered = false;

    if (u.isLocalFile()) {
        // the empty path is kind of a hack
        KFileViewItem i("", u.path());
        if (i.isDir())
            setURL(text, true);
        else {
            if (takeFiles)
                if (acceptOnlyExisting && !i.isFile())
                    warning("you entered an invalid URL");
                else
                    filenameEntered = true;
        }
    } else
        setURL(text, true);

    if (filenameEntered) {
        filename_ = u.url();
        emit fileSelected(filename_);

        if (!finished)
            QApplication::restoreOverrideCursor();
        finished = false;

        accept();
    }
#endif
//    kdDebug() << "TODO DirOperator::checkPath()" << endl;
}

void DirOperator::setURL(const KURL& _newurl, bool clearforward)
{
    KURL newurl = _newurl;

    QString pathstr = newurl.path(+1);

    // already done by the +1 above
    //if (pathstr.isEmpty() || pathstr.at(pathstr.length() - 1) != '/')
    //    pathstr += '/';
    newurl.setPath(pathstr);


    if (newurl == dir->url()) // already set
        return;



    pendingMimeTypes.clear();

    if (clearforward) {
        // autodelete should remove this one
        backStack.push(new KURL(dir->url()));
        forwardStack.clear();
    }
    /* // FIXME: (pfeiffer) we should have a flag "onlyLocal", I guess
    KURL testURL(pathstr);
    if ( !testURL.isLocalFile() ) {
        KMessageBox::sorry(0,
                           i18n("The specified directory is not a "
                                "local directory\n"
                                "But the application accepts just "
                                "local files."));
        return;
    }
    */

    KURL backup(dir->url());
    dir->setURL(newurl);

    bool ok = false;
    if (!dir->isReadable()) {
        // maybe newurl is a file? check its parent directory
        newurl.cd(QString::fromLatin1(".."));
        dir->setURL(newurl);

        if ( !dir->isReadable() ) {
            KMessageBox::error(0,
                               i18n("The specified directory does not exist\n"
                                    "or was not readable."));
            dir->setURL(backup);
            backStack.pop();
        }
        else
            // don't reread when the url stayed the same
            ok = (dir->url().url(-1) != newurl.url(-1));
    }
    else
        ok = true;

    if ( ok ) {
        myCompletion.clear();
        myDirCompletion.clear();
        emit urlEntered(dir->url());
        pathChanged();
    }


    // enable/disable actions
    forwardAction->setEnabled( !forwardStack.isEmpty() );
    backAction->setEnabled( !backStack.isEmpty() );
    upAction->setEnabled( !isRoot() );
}

void DirOperator::rereadDir()
{
    pathChanged();
}

// Protected
void DirOperator::pathChanged()
{
    if (!fileView)
        return;

    pendingMimeTypes.clear();
    fileView->clear();
    myCompletion.clear();
    myDirCompletion.clear();

    // it may be, that we weren't ready at this time
    if (!finished)
        QApplication::restoreOverrideCursor();

    finished = false;

    // when KIO::Job emits finished, the slot will restore the cursor
    QApplication::setOverrideCursor( waitCursor );

    // lastDirectory is used to set the start path next time
    // we select a file
    *lastDirectory = dir->url();

    if (!dir->isReadable()) {
        KMessageBox::error(0,
                           i18n("The specified directory does not exist "
                                "or was not readable."));
        if (backStack.isEmpty())
            home();
        else
            back();
    }

    dir->listDirectory();
}

// Code pinched from kfm then hacked
void DirOperator::back()
{
    if ( backStack.isEmpty() )
        return;

    forwardStack.push( new KURL(dir->url()) );

    KURL *s = backStack.pop();

    setURL(*s, false);
    delete s;
}

// Code pinched from kfm then hacked
void DirOperator::forward()
{
    if ( forwardStack.isEmpty() )
        return;

    backStack.push(new KURL(dir->url()));

    KURL *s = forwardStack.pop();
    setURL(*s, false);
    delete s;
}

KURL DirOperator::url() const
{
    return dir->url();
}

void DirOperator::cdUp()
{
    KURL tmp( dir->url() );
    tmp.cd(QString::fromLatin1(".."));
    setURL(tmp, true);
}

void DirOperator::home()
{
    setURL(QDir::homeDirPath(), true);
}

void DirOperator::setNameFilter(const QString& filter)
{
    dir->setNameFilter(filter);
}

bool DirOperator::isRoot() const
{
    return dir->isRoot();
}

void DirOperator::setView( KFile::FileView view,bool mod )
{
    bool separateDirs = KFile::isSeparateDirs( view );

    if (view == KFile::Default) {
        if ( KFile::isDetailView( (KFile::FileView) defaultView ) )
            view = KFile::Detail;
        else
            view = KFile::Simple;

        if(mod) separateDirs = KFile::isSeparateDirs( (KFile::FileView) defaultView );
        else separateDirs = true;
    }

    bool preview=( (view & KFile::PreviewInfo) == KFile::PreviewInfo ||
                   (view & KFile::PreviewContents) == KFile::PreviewContents );

    // we only have a dual combi view, not a triple one. So in Directory and
    // preview mode, we don't allow separating dirs & files
    if ( (mode() & KFile::Directory) == KFile::Directory || preview )
        separateDirs = false;

    viewKind = static_cast<int>(view) | (separateDirs ? KFile::SeparateDirs : 0);

    KFileView *new_view = 0L;
    if (separateDirs) {
        KCombiView *combi = new KCombiView(this, "combi view");
        combi->setOnlyDoubleClickSelectsFiles(d->onlyDoubleClickSelectsFiles);
        new_view = combi;

        if ( (view & KFile::Simple) == KFile::Simple ) {
            KFileDnDIconView *view = new KFileDnDIconView( combi, "simple view" );
    connect(view,SIGNAL(dropped(QStringList)),SLOT(dropps(QStringList)));
            view->setViewName( i18n("Short View") );
            combi->setRight( view );
        }
        else
            combi->setRight(new KFileDndDetailView( combi, "detail view" ));
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

    } else {
        if ( (view & KFile::Simple) == KFile::Simple && !preview ) {
            new_view = new KFileDnDIconView( this, "simple view" );
            new_view->setViewName( i18n("Short View") );
				    connect(static_cast<KFileDnDIconView*>(new_view),SIGNAL(dropped(QStringList)),this,SLOT(dropps(QStringList)));
        }
        else if ( (view & KFile::Detail) == KFile::Detail && !preview )	{
            new_view = new KFileDndDetailView( this, "detail view" );
				    connect(static_cast<KFileDndDetailView*>(new_view),SIGNAL(dropped(QStringList)),SLOT(dropps(QStringList)));		
				    	}
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        else { // FIXME: make Preview and Detail possible?
            KFilePreview *tmp = new KFilePreview(this, "preview");
           tmp->setOnlyDoubleClickSelectsFiles(d->onlyDoubleClickSelectsFiles);
            tmp->setPreviewWidget(myPreview, url());
            new_view=tmp;
        }
    }

    if ( (mode() & KFile::Directory) )
        new_view->setViewMode(KFileView::Directories);
    else
        new_view->setViewMode(KFileView::All);

    setFocusProxy(new_view->widget());
    new_view->setSorting( mySorting );
    new_view->setOnlyDoubleClickSelectsFiles( d->onlyDoubleClickSelectsFiles );
    connectView(new_view);
}

void DirOperator::deleteOldView()
{
    if ( oldView ) {
        oldView->widget()->removeEventFilter( this );
        oldView->widget()->hide();
        delete oldView;
        oldView = 0;
    }
}


void DirOperator::connectView(KFileView *view)
{
    // TODO: do a real timer and restart it after that
    pendingMimeTypes.clear();

    if (fileView) {
        QApplication::setOverrideCursor( waitCursor );
        finished = false;
        fileView->setOperator(0);
        // it's _very_ unlikly that oldView still has a value
        oldView = fileView;
        // the timer didn't work (KFileItems being deleted, then the view
        // being deleted, the view deleting the view items, which tried to
        // access the (dead) KFileItems. What was it good for anyway?
        // QTimer::singleShot(0, this, SLOT(deleteOldView()));
        deleteOldView();
    }

    fileView = view;
    fileView->setOperator(this);
    if ( reverseAction->isChecked() != fileView->isReversed() )
        fileView->sortReversed();

    if ( (myMode & KFile::File) == KFile::File )
    	{
        fileView->setViewMode( KFileView::All );
        fileView->setSelectionMode( KFile::Single);//s );
    	}
    else if ( (myMode & KFile::Directory) == KFile::Directory )
    	{
        fileView->setViewMode( KFileView::Directories );
        fileView->setSelectionMode( KFile::Single);//s );
//        fileView->setSelectionMode( KFile::Extended);//s );
    	}
    else if ( (myMode & KFile::Files) == KFile::Files )
    {
        fileView->setViewMode( KFileView::All );
        fileView->setSelectionMode( KFile::Extended );
    }
    else if ( (viewKind & KFile::PreviewContents) == KFile::PreviewContents ||
              (viewKind & KFile::PreviewInfo) == KFile::PreviewInfo )
    	{
        fileView->setViewMode( KFileView::All );
        fileView->setSelectionMode( KFile::Single);//s );
    	}

    dir->listDirectory();

    updateViewActions();
    fileView->widget()->show();
    fileView->widget()->resize(size());

    // FIXME : only installed because focusNextPrevChild() isn't called
    fileView->widget()->setAcceptDrops(true);
    fileView->widget()->installEventFilter( this );
		

//    connect(view->widget(),SIGNAL(dropped()),SLOT(drops()));

}

KFile::Mode DirOperator::mode() const
{
    return myMode;
}

void DirOperator::setMode(KFile::Mode m)
{
    if (myMode == m)
        return;

    myMode = m;

    dir->setDirOnlyMode( ( myMode == KFile::Directory ) );

    // reset the view with the different mode
    setView( static_cast<KFile::FileView>(viewKind) );
}

void DirOperator::setView(KFileView *view,bool mod)
{
	/* temporary */ mod = false;
    setFocusProxy(view->widget());
    view->setSorting( mySorting );
    connectView(view);
}

void DirOperator::setFileReader( FileReader *reader )
{
    delete dir;
    dir = reader;
}

// files from a remote url will be added immediately, but won't get any
// mimetype-detection
// for local files, we will wait until the entire loading is finished and
// then dump them into the view in one chunk
void DirOperator::insertNewFiles(const KFileItemList &newone)
{
    if (newone.isEmpty())
        return;

    myCompleteListDirty = true;
    bool isLocal = dir->url().isLocalFile();
    if ( !isLocal )
        insertIntoView(newone);

    if (!dir->isFinished()) {
        if (!progress) {
            progress = new KProgress(this, "progress");
            progress->adjustSize();
            progress->setRange(0, 100);
            progress->move(2, height() - progress->height() -2);
        }

        progress->setValue(ulong(dir->items().count() * 100)/ dir->items().count());
        progress->raise();
        progress->show();

        // we have to redraw this in as fast as possible
        QApplication::flushX();
    }
}

void DirOperator::selectDir(const KFileViewItem *item)
{
    KURL tmp( dir->url() );
    tmp.cd(item->name());
    setURL(tmp, true);
//    setURL(item->url(), true);
}

void DirOperator::itemDeleted(KFileItem *item)
{
    fileView->removeItem( static_cast<KFileViewItem *>( item ));
    emit updateInformation(fileView->numDirs(), fileView->numFiles());
}

void DirOperator::selectFile(const KFileViewItem *item)
{
    // FIXME: what is this about?? (pfeiffer)
    KURL tmp ( dir->url() );
    tmp.setFileName(item->name());

    if (!finished)
        QApplication::restoreOverrideCursor();
    finished = false;

    emit fileSelected( item );
}


void DirOperator::filterChanged()
{
    pendingMimeTypes.clear();
    fileView->clear();
    myCompletion.clear();
    myDirCompletion.clear();
    dir->listDirectory();
}

void DirOperator::setCurrentItem( const QString& filename )
{
    const KFileViewItem *item = 0L;

    if ( !filename.isNull() )
        item = static_cast<KFileViewItem *>(dir->findByName( filename ));

    fileView->clearSelection();
    if ( item )
        fileView->setCurrentItem( QString::null, item );
}

QString DirOperator::makeCompletion(const QString& string)
{
    if ( string.isEmpty() ) {
        fileView->clearSelection();
        return QString::null;
    }

    prepareCompletionObjects();
    return myCompletion.makeCompletion( string );
}

QString DirOperator::makeDirCompletion(const QString& string)
{
    if ( string.isEmpty() ) {
        fileView->clearSelection();
        return QString::null;
    }

    prepareCompletionObjects();
    return myDirCompletion.makeCompletion( string );
}

void DirOperator::prepareCompletionObjects()
{
    if ( myCompleteListDirty ) { // create the list of all possible completions
        KFileItemListIterator it( dir->items());
        for( ; it.current(); ++it ) {
            KFileItem *item = it.current();

            myCompletion.addItem( item->name() );
            if ( item->isDir() )
                myDirCompletion.addItem( item->name() );
        }
        myCompleteListDirty = false;
    }
}

void DirOperator::slotCompletionMatch(const QString& match)
{
    setCurrentItem( match );
    emit completion( match );
}

void DirOperator::setupActions()
{
    actionMenu = new KActionMenu( i18n("Menu"), this, "popupMenu" );

    upAction = KStdAction::up( this, SLOT( cdUp() ), this, "up" );
    backAction = KStdAction::back( this, SLOT( back() ), this, "back" );
    forwardAction = KStdAction::forward(this, SLOT(forward()), this,"forward");
    homeAction = KStdAction::home( this, SLOT( home() ), this, "home" );
    homeAction->setText(i18n("Home directory"));
    reloadAction =KStdAction::redisplay(this,SLOT(rereadDir()),this, "reload");
    actionSeparator = new KActionSeparator( this, "separator" );
    mkdirAction = new KAction( i18n("New Folder..."), 0, this, SLOT( mkdir() ), this, "mkdir");
    mkdirAction->setIcon( QString::fromLatin1("folder_new") );
    renFileAction = new KAction( i18n("Rename..."), 0, this, SLOT( rename() ), this, "rename");
    renFileAction->setIcon( QString::fromLatin1("reload") );
    openWithAction = new KAction( i18n("Open with..."), 0, this, SLOT( openWith() ), this, "openWith");
    openWithAction->setIcon( QString::fromLatin1("misc") );
    bookmarkAction = new KAction( i18n("Bookmark current path..."), 0, this, SLOT( bookMark() ), this, "bookmark");
    bookmarkAction->setIcon( QString::fromLatin1("bookmark_add") );
    setFilterAction = new KAction( i18n("set filter to..."), 0, this, SLOT( setFilter() ), this, "filter");
    setFilterAction->setIcon( QString::fromLatin1("filter") );
#ifdef MAIL
    emailAction = new KAction( i18n("Send file as Mail"), 0, this, SLOT( sendMail() ), this, "email");
    emailAction->setIcon( QString::fromLatin1("kmail") );
#endif

    // FIXME, find a better icon than filenew
    reloadAction->setText( i18n("Reload") );


    // the sort menu actions
    sortActionMenu = new KActionMenu( i18n("Sorting"), this, "sorting menu");
    byNameAction = new KRadioAction( i18n("By Name"), 0, this, SLOT( slotSortByName() ), this, "by name" );
    byDateAction = new KRadioAction( i18n("By Date"), 0, this, SLOT( slotSortByDate() ), this, "by date" );
    bySizeAction = new KRadioAction( i18n("By Size"), 0, this, SLOT( slotSortBySize() ), this, "by size" );
    reverseAction = new KToggleAction( i18n("Reverse"), 0, this, SLOT( slotSortReversed() ), this, "reversed" );

    QString sortGroup = QString::fromLatin1("sort");
    byNameAction->setExclusiveGroup( sortGroup );
    byDateAction->setExclusiveGroup( sortGroup );
    bySizeAction->setExclusiveGroup( sortGroup );


    dirsFirstAction = new KToggleAction( i18n("Directories first"), 0, this, "dirs first");
    caseInsensitiveAction = new KToggleAction(i18n("Case insensitive"), 0, this, "case insensitive" );

    connect( dirsFirstAction, SIGNAL( toggled( bool ) ), SLOT( slotToggleDirsFirst() ));
    connect( caseInsensitiveAction, SIGNAL( toggled( bool ) ), SLOT( slotToggleIgnoreCase() ));



    // the view menu actions
    viewActionMenu = new KActionMenu( i18n("View"), this, "view menu" );
    shortAction = new KRadioAction( i18n("Short View"), 0, this, "short view" );
    detailedAction = new KRadioAction( i18n("Detailed View"), 0, this, "detailed view" );

    showHiddenAction = new KToggleAction( i18n("Show Hidden Files"), 0, this, "show hidden" );
    separateDirsAction = new KToggleAction( i18n("Separate Directories"), 0, this, SLOT(slotToggleMixDirsAndFiles()), this, "separate dirs" );

    QString viewGroup = QString::fromLatin1("view");
    shortAction->setExclusiveGroup( viewGroup );
    detailedAction->setExclusiveGroup( viewGroup );

    connect( shortAction, SIGNAL( toggled( bool ) ),
             SLOT( slotSimpleView() ));
    connect( detailedAction, SIGNAL( toggled( bool ) ),
             SLOT( slotDetailedView() ));
    connect( showHiddenAction, SIGNAL( toggled( bool ) ),
             SLOT( slotToggleHidden( bool ) ));


    // insert them into the actionCollection
    myActionCollection = new KActionCollection( this, "action collection" );
    myActionCollection->insert( actionMenu );
    myActionCollection->insert( backAction );
    myActionCollection->insert( forwardAction );
    myActionCollection->insert( homeAction );
    myActionCollection->insert( upAction );
    myActionCollection->insert( reloadAction );
    myActionCollection->insert( actionSeparator );
    myActionCollection->insert( mkdirAction );
    myActionCollection->insert( renFileAction );
    myActionCollection->insert( openWithAction );
    myActionCollection->insert( bookmarkAction );


    // Start FUX
#ifdef MAIL
    myActionCollection->insert( emailAction );
#endif
    // End FUX


    myActionCollection->insert( setFilterAction );

    myActionCollection->insert( sortActionMenu );
    myActionCollection->insert( byNameAction );
    myActionCollection->insert( byDateAction );
    myActionCollection->insert( bySizeAction );
    myActionCollection->insert( reverseAction );
    myActionCollection->insert( dirsFirstAction );
    myActionCollection->insert( caseInsensitiveAction );
    myActionCollection->insert( viewActionMenu );
    myActionCollection->insert( shortAction );
    myActionCollection->insert( detailedAction );
    myActionCollection->insert( showHiddenAction );
    myActionCollection->insert( separateDirsAction );
}


void DirOperator::setupMenu()
{
    // first fill the submenus (sort and view)
    sortActionMenu->insert( byNameAction );
    sortActionMenu->insert( byDateAction );
    sortActionMenu->insert( bySizeAction );
    sortActionMenu->insert( reverseAction );
    sortActionMenu->insert( actionSeparator );
    sortActionMenu->insert( dirsFirstAction );
    sortActionMenu->insert( caseInsensitiveAction );

    viewActionMenu->insert( shortAction );
    viewActionMenu->insert( detailedAction );
    viewActionMenu->insert( actionSeparator );
    viewActionMenu->insert( showHiddenAction );
    viewActionMenu->insert( separateDirsAction );


    // now plug everything into the popupmenu
    actionMenu->insert( upAction );
    actionMenu->insert( backAction );
    actionMenu->insert( forwardAction );
    actionMenu->insert( homeAction );
    actionMenu->insert( actionSeparator );

    actionMenu->insert( mkdirAction );
    actionMenu->insert( renFileAction );
    actionMenu->insert( openWithAction );
    actionMenu->insert( bookmarkAction );
    actionMenu->insert( setFilterAction );
    // FUX
#ifdef MAIL
    actionMenu->insert( emailAction );
#endif

    actionMenu->insert( actionSeparator );

    actionMenu->insert( sortActionMenu );
    actionMenu->insert( actionSeparator );
    actionMenu->insert( viewActionMenu );
}


void DirOperator::updateSortActions()
{
    if ( KFile::isSortByName( mySorting ) )
        byNameAction->setChecked( true );
    else if ( KFile::isSortByDate( mySorting ) )
        byDateAction->setChecked( true );
    else if ( KFile::isSortBySize( mySorting ) )
        bySizeAction->setChecked( true );

    dirsFirstAction->setChecked( KFile::isSortDirsFirst( mySorting ));
    caseInsensitiveAction->setChecked(KFile::isSortCaseInsensitive(mySorting));
}

void DirOperator::updateViewActions()
{
    KFile::FileView fv = static_cast<KFile::FileView>( viewKind );

    separateDirsAction->setChecked( KFile::isSeparateDirs( fv ));

    shortAction->setChecked( KFile::isSimpleView( fv ));
    detailedAction->setChecked( KFile::isDetailView( fv ));
}


void DirOperator::readConfig( KConfig *kc, const QString& group )
{
    if ( !kc )
        return;
    QString oldGroup = kc->group();
    if ( !group.isEmpty() )
        kc->setGroup( group );

    defaultView = 0;
    int sorting = 0;

    QString viewStyle = kc->readEntry( QString::fromLatin1("View Style"),
                                       QString::fromLatin1("Simple") );
    if ( viewStyle == QString::fromLatin1("Detail") )
        defaultView |= KFile::Detail;
    else
        defaultView |= KFile::Simple;
    if ( kc->readBoolEntry( QString::fromLatin1("Separate Directories"),
                             DefaultMixDirsAndFiles ) )
        defaultView |= KFile::SeparateDirs;
    if ( kc->readBoolEntry( QString::fromLatin1("Sort case insensitively"),
                            DefaultCaseInsensitive ) )
        sorting |= QDir::IgnoreCase;
    if ( kc->readBoolEntry( QString::fromLatin1("Sort directories first"),
                            DefaultDirsFirst ) )
        sorting |= QDir::DirsFirst;


    QString name = QString::fromLatin1("Name");
    QString sortBy = kc->readEntry( QString::fromLatin1("Sort by"), name );
    if ( sortBy == name )
        sorting |= QDir::Name;
    else if ( sortBy == QString::fromLatin1("Size") )
        sorting |= QDir::Size;
    else if ( sortBy == QString::fromLatin1("Date") )
        sorting |= QDir::Time;

    mySorting = static_cast<QDir::SortSpec>( sorting );
    setSorting( mySorting );


    if ( kc->readBoolEntry( QString::fromLatin1("Show hidden files"),
                            DefaultShowHidden ) ) {
         showHiddenAction->setChecked( true );
         dir->setShowingDotFiles( true );
    }
    if ( kc->readBoolEntry( QString::fromLatin1("Sort reversed"),
                            DefaultSortReversed ) )
        reverseAction->setChecked( true );

    kc->setGroup( oldGroup );
}

void DirOperator::saveConfig( KConfig *kc, const QString& group )
{
    if ( !kc )
        return;

    const QString oldGroup = kc->group();

    if ( !group.isEmpty() )
        kc->setGroup( group );

    QString sortBy = QString::fromLatin1("Name");
    if ( KFile::isSortBySize( mySorting ) )
        sortBy = QString::fromLatin1("Size");
    else if ( KFile::isSortByDate( mySorting ) )
        sortBy = QString::fromLatin1("Date");
    kc->writeEntry( QString::fromLatin1("Sort by"), sortBy );

    kc->writeEntry( QString::fromLatin1("Sort reversed"),
                    reverseAction->isChecked() );
    kc->writeEntry( QString::fromLatin1("Sort case insensitively"),
                    caseInsensitiveAction->isChecked() );
    kc->writeEntry( QString::fromLatin1("Sort directories first"),
                    dirsFirstAction->isChecked() );

    kc->writeEntry( QString::fromLatin1("Separate Directories"),
                    separateDirsAction->isChecked() );
    kc->writeEntry( QString::fromLatin1("Show hidden files"),
                    showHiddenAction->isChecked() );

    KFile::FileView fv = static_cast<KFile::FileView>( viewKind );
    QString style;
    if ( KFile::isDetailView( fv ) )
        style = QString::fromLatin1("Detail");
    else if ( KFile::isSimpleView( fv ) )
        style = QString::fromLatin1("Simple");
    kc->writeEntry( QString::fromLatin1("View Style"), style );

    kc->setGroup( oldGroup );
}

void DirOperator::setDirsExtra()
{
}

void DirOperator::resizeEvent( QResizeEvent * )
{
    if (fileView)
        fileView->widget()->resize( size() );
    if ( progress )
        progress->move(2, height() - progress->height() -2);
}

void DirOperator::setOnlyDoubleClickSelectsFiles( bool enable )
{
    d->onlyDoubleClickSelectsFiles = enable;
}

bool DirOperator::onlyDoubleClickSelectsFiles() const
{
    return d->onlyDoubleClickSelectsFiles;
}
//ARNO///////////////////////////////////////////////////////////////////////ARNO////////////////
void DirOperator::insertIntoView(const KFileItemList& items)
{
    if ( items.isEmpty() )
        return;

    pendingMimeTypes.clear();
    KFileViewItemList list;
    KFileItemListIterator it( items );
    register bool isLocal = dir->url().isLocalFile();

    while ( it.current() ) {
        list.append( static_cast<KFileViewItem *>( it.current() ));
        if ( isLocal )
            pendingMimeTypes.append(static_cast<KFileViewItem*>(it.current()));
        ++it;
    }
    fileView->addItemList( list );
    emit updateInformation(fileView->numDirs(), fileView->numFiles());
}
////////////////////////////////////////////////////////////////////////////////////////////
// local files will be inserted in one big chunk
void DirOperator::slotIOFinished()
{
    // this sucks currently, as this slot is also called by KDirLister after
    // an update, where we don't want to clear the list -- FIXME
    if ( dir->url().isLocalFile() ) {
        fileView->clear();
        insertIntoView( dir->items() );
    }

    QTimer::singleShot(0, this, SLOT(readNextMimeType()));
    QTimer::singleShot(200, this, SLOT(resetCursor()));
    emit finishedLoading();
}
////////////////////////////////////////////////////////////////////////////////////////////

// somehow focusNextPrevChild() is not called ;(
bool DirOperator::eventFilter( QObject *o, QEvent *e )
{
    if(e->type() == QEvent::FocusIn ) emit hatFocus();
    if(e->type() == QEvent::FocusIn ) emit hatKeinFocus();
    if(e->type() == QEvent::KeyPress) keyPressEvent((QKeyEvent*)e);
    return QWidget::eventFilter( o, e );
}

void DirOperator::keyPressEvent(QKeyEvent *e)
{
	emit keyPress(e);
	e->ignore();
}
	
KFileItem* DirOperator::findByName( const QString& name ) const
{
	return dir->findByName(name);
}

void DirOperator::dropps(QStringList list)
{
	emit droppedFiles(list);
}

void DirOperator::rename()
{
	emit renameFile();
}

void DirOperator::openWith()
{
	emit openWithSignal();
}

void DirOperator::bookMark()
{
	KConfig *config = kapp->config();
	config->setGroup("Assignments");
	QString curPath = url().path();
	
  QString l,s,t;
  int i;
  for(i = 'A'; i <= 'Z';i++)
  {
      l.sprintf("%c",i);
			t.sprintf("DriveAssign%c",i);	
			s = config->readEntry(t,"");
 			if(s == "") break;
  }
	s.sprintf("DriveAssign%c",i);
	config->writeEntry(s,curPath);
}

void DirOperator::setFilter()
{
	InputBox *ip = new InputBox(this,"input filtername","*.*");
	ip->exec();
	QString temp = ip->returnInput();
	delete ip;
  setNameFilter(temp);

// Refresh of view !!
	KURL u1 = url();
	cdUp();
	setURL(u1,false); // It's not the best but fast enough hehe .. we have to implement an refreshprocedure !!!!!!
}

void DirOperator::setSelection(QString selString)
{
	if (selString == "*.*") {
		fileView->selectAll();
		return;
	}
	QString tempS = selString.right(selString.length()-1);

  const	KFileViewItemList *item = fileView->items();
  QListIterator<KFileViewItem> it(*item);
		for ( ; it.current(); ++it ) {
			KFileViewItem *item = it.current();
			if(item->name().contains(tempS)) fileView->setSelected(item,true);	
		}
}

void DirOperator::setUnSelection(QString selString)
{
	if (selString == "*.*") {
		fileView->selectAll();
		return;
	}
	QString tempS = selString.right(selString.length()-1);

  const	KFileViewItemList *item = fileView->items();
  QListIterator<KFileViewItem> it(*item);
		for ( ; it.current(); ++it ) {
			KFileViewItem *item = it.current();
			if(item->name().contains(tempS)) fileView->setSelected(item,false);	
		}
}

void DirOperator::setSelectionAll()
{
	fileView->selectAll();
}

void DirOperator::setSelectionInvert()
{
	fileView->invertSelection();
}

void DirOperator::setSelectionNone()
{
	fileView->clearSelection();
}
void DirOperator::sendMail()
{
	emit sendFile();
}
